/******************************************************************************
 * $Header: /CommonBe/agmsmith/Programming/AGMSRAMFileSystem/RCS/AVLDupTree.c,v 1.23 2003/08/10 19:02:21 agmsmith Exp $
 *
 * This set of C subroutines is useful for indexing a set of key/value pairs,
 * using the key to find a matching value.  As an enhancement, multiple values
 * for the same key are supported efficiently.
 *
 * Copyright © 2001 by Alexander G. M. Smith.  Look a bit furthur down in this
 * file for licensing details (GNU Lesser General Public License).
 *
 * It's implemented with a balanced binary search tree, which lets you store
 * and find things in optimal O(log n) time.  A binary search tree (BST) has
 * the property that all nodes in the left sub-tree have keys less than the
 * given node, and all nodes in the right sub-tree have keys greater than the
 * given node.  Searching is done by just traversing down and choosing left or
 * right depending if the key you are looking for is less than or greater than
 * the current node's key.  The balancing is done by using the AVL algorithm,
 * which enforces the condition that left and right sub-trees differ in height
 * by at most one.  If an insert or delete operation (done much like a search -
 * except that you add the new node where the search said it should be) breaks
 * that rule, the subtrees are fixed up with a "BST rotation" operation, with
 * recursive fixups applied from the point of change all the way to the root if
 * necessary.  Since finding a key/value pair and the number of fixups for
 * insert/delete both depend on the height of the tree (which is on the order
 * of the natural logarithm (actually the base 2 logarithm is more appropriate
 * but natural is just a multiple of it) of the number of items (n) in the
 * tree, or in other notation: O(log n)), and a single fixup or search
 * comparision takes constant time, the overall speed of the basic
 * find/insert/delete operations is O(log n).
 *
 * The AVL tree algorithms were invented by G. M. Adel'son-Velskii and
 * Y. M. Landis, who wrote about it first in 1962 in Soviet Math Dockl., titled
 * "An Algorithm for the Organization of Information".  I made use of a nice
 * article written by Timothy Rolfe on AVL trees in the December 2000 issue of
 * Dr. Dobb's Journal, pages 149-152.  I also dug up an alternate description
 * of AVL trees using "left heavy" and "right heavy" concepts in an old
 * textbook called "An Introduction to Data Structures with Applications", by
 * Jean-Paul Tremblay and P. G. Sorenson, 1976, McGraw-Hill computer science
 * series, pages 489-500.  They have a nice reference on page 519 to
 * Adel'son-Vel'skii, G. M. and E. M. Landis, same title as before, in
 * "Dokl. Akad. Nauk SSSR, Mathemat., vol. 146, no. 2, pp. 263-266, 1962."
 *
 * It's called AVLDupTree since it has been altered by Alexander G. M. Smith
 * to handle duplicate keys efficiently.  That was originally done by making a
 * subtree under the node with the common key to store the values.  But that
 * turned out to be unnecessarily complicated, just changing the key comparison
 * function solves the same problem - it now looks at the values if the keys
 * were equal.
 *
 * Since this is designed for use as the index for a BeOS inspired file system,
 * the keys can be the typical BeOS keys: string, int32, int64, float, double.
 * Because the AGMSRAMFileSystem doesn't need arbitrary data types as the
 * values associated with a key, values are just an unsigned int (same size as
 * a pointer).
 *
 * It is not designed to work in a multitasking environment (the more general
 * version of this library is), the caller must avoid accessing the tree from
 * different threads (read only is OK, but if any thread is changing the
 * thread, all other threads must be prevented from using the tree data
 * structures).
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation; either version 2.1 of the License, or (at your
 * option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 * for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Send comments and wishes to agmsmith@rogers.com (or do a web search for
 * "Alexander G. M. Smith" to find a more recent e-mail address and web pages).
 *
 * $Log: AVLDupTree.c,v $
 * Revision 1.23  2003/08/10 19:02:21  agmsmith
 * Changed time to be a signed 32 bit integer.
 *
 * Revision 1.22  2003/04/27 00:03:21  agmsmith
 * Add a unique code after the fake name used in indexes when seen as
 * directories, so that several entries in the index that have the same
 * key (such as files the same size in the size index) appear to be
 * different to old programs.  Also need to check if a particular key
 * and value are in the AVL tree.
 *
 * Revision 1.21  2003/04/13 21:09:45  agmsmith
 * Time is a 32 bit value, UNIX style.
 *
 * Revision 1.20  2003/01/12 00:51:57  agmsmith
 * Added external AVLDupGetComparisonFunctionForType function.
 *
 * Revision 1.19  2003/01/11 22:48:56  agmsmith
 * Allow a NULL for the allocation and deallocation function pointers.
 *
 * Revision 1.18  2002/07/02 20:52:44  agmsmith
 * Oops.
 *
 * Revision 1.17  2002/07/02 20:45:39  agmsmith
 * MIME strings are B_MIME_STRING_TYPE, not B_MIME_TYPE (whatever that is).
 *
 * Revision 1.16  2002/06/29 21:38:36  agmsmith
 * Added AVLDupGetThingTypeSize.
 *
 * Revision 1.15  2002/06/12 19:39:33  agmsmith
 * Added plus and minus infinity value test option for iteration.
 *
 * Revision 1.14  2002/06/06 22:34:56  agmsmith
 * Change comparison to avoid number overflow when comparing values
 * while iterating (uint comparison directly rather than subtracting).
 *
 * Revision 1.13  2002/06/06 16:53:42  agmsmith
 * Can't use parsedate, instead just use simple date format of the
 * number of seconds since the start of time.
 *
 * Revision 1.12  2002/06/04 20:47:54  agmsmith
 * Under construction, looking for date functions in kernel land.
 *
 * Revision 1.11  2002/05/31 23:28:31  agmsmith
 * Added a bunch of new data types for the key.
 *
 * Revision 1.10  2002/01/31 21:53:18  agmsmith
 * E-mail address updated.
 *
 * Revision 1.9  2002/01/31 21:50:59  agmsmith
 * Changed values to be unsigned ints rather than void *, also
 * added AVLDupChangeValueInPlace and had to change the
 * iteration callback to use a pointer to the value not the value.
 *
 * Revision 1.8  2001/11/20 19:13:38  agmsmith
 * Added an error code for bad arguments when adding to the tree.
 *
 * Revision 1.7  2001/10/16 22:05:17  agmsmith
 * Changed return code from Add to specify separate codes for
 * adding a new node and trying to add an existing key/value.
 *
 * Revision 1.6  2001/07/29 20:06:05  agmsmith
 * Added a backwards iteration option and a FindPrevious function.
 *
 * Revision 1.5  2001/07/23 01:11:24  agmsmith
 * Added utility functions for quickly finding a key, basically
 * just a wrapper around the generic search function.
 *
 * Revision 1.4  2001/07/21 22:58:57  agmsmith
 * PoolID should be a const void *.
 *
 * Revision 1.3  2001/07/20 01:43:25  agmsmith
 * Cleaned up include files to only use necessary ones.
 *
 * Revision 1.2  2001/07/16 20:14:00  agmsmith
 * Changed to use memory pools, and have values as a (void *) rather than
 * a generic data type.  Also multitasking semaphores removed and a
 * few other frills taken out.
 *
 * Revision 1.1  2001/07/06 00:16:55  agmsmith
 * Initial revision
 *
 * Revision 1.11  2001/03/28 22:17:44  agmsmith
 * Prepared for 1.0 release.
 *
 * Revision 1.10  2001/03/23 02:41:30  agmsmith
 * Oops...
 *
 * Revision 1.9  2001/03/22 22:58:13  agmsmith
 * Added double bounded iteration function, compiles but not tested yet.
 *
 * Revision 1.8  2001/03/19 01:43:51  agmsmith
 * Pass largely unchanging arguments using a per-thread structure,
 * rather than having separate arguments in each recursive call.
 *
 * Revision 1.7  2001/03/18 17:53:23  agmsmith
 * Added semaphore for multitasking safety.
 *
 * Revision 1.6  2001/03/17 21:09:01  agmsmith
 * Printed it out, read it over, corrected a few mistakes.
 *
 * Revision 1.5  2001/03/16 22:31:30  agmsmith
 * Added AVL balancing, and implemented delete.
 *
 * Revision 1.4  2001/03/15 18:02:13  agmsmith
 * Removed duplicates tree, partially working now without it.
 *
 * Revision 1.3  2001/03/12 22:38:56  agmsmith
 * Just before removing duplicates subtree, but it does work a bit.
 *
 * Revision 1.2  2001/03/10 17:59:19  agmsmith
 * Under construction.
 *
 * Revision 1.1  2001/03/08 20:36:31  agmsmith
 * Initial revision
 */

#include <TypeConstants.h> /* For type enum items like B_STRING_TYPE. */
#include <stdlib.h>        /* For atol and similar conversion functions. */
#include <stdio.h>         /* For sprintf. */
#include <string.h>        /* For strncpy. */
#define B_MIME_STRING_TYPE  'MIMS' /* Since we can't use C++ish Mime.h */

#include "AVLDupTree.h"


/* Force an error if someone attempts to use the standard malloc and free,
should be using memory pools instead. */

#define malloc bleeble
#define free bleeble


/* This structure is a node in the AVL tree.  The keys in the sub-tree at
smallerChildPntr are all less than the node's key.  Keys in the sub-tree rooted
at largerChildPntr are all larger than the node's key.  The height is the
maximum of the height of the smaller and greater children, plus 1. A node with
no children has a height of 1, similarly a NULL pointer (empty tree) has a
height of 0. To keep the tree balanced, nodes are moved around so that the
heights of the smaller and larger children differ by at most 1. */

typedef struct AVLDupNodeStruct AVLDupNodeRecord, *AVLDupNodePointer;

struct AVLDupNodeStruct
{
  AVLDupThingRecord key;
  unsigned int      value;
  AVLDupNodePointer smallerChildPntr;
  AVLDupNodePointer largerChildPntr;
  unsigned int      height; /* Only needs to be uint8, rest is for padding. */
};


/* The header for the AVLDupTree itself.  Besides pointing out the root node,
it contains auxiliary information needed for comparing the associated data
types, for collecting statistics, and for doing pool memory allocation. */

struct AVLDupTreeStruct
{
  type_code keyType;
  AVLDupComparisonFunctionPointer keyComparisonFunctionPntr;
  AVLDupNodePointer rootPntr; /* Root node of the tree or NULL. */
  unsigned int count; /* Counts user provided key/value pairs in tree. */
  void * (*mallocPntr) (const void *pool_id, size_t size);
  void (*freePntr) (const void *pool_id, void *ptr);
  const void *poolID;
};


/* This structure is used for keeping common data around during recursive
function calls.  It is faster than the simple technique of having a separate
argument for everything the recursive function might need (like the key and
value you are searching for) - only a single pointer to this structure is
passed into the recursive routines rather than many parameters.  It's also
safer than using global variables, which limit multitasking access to the tree
to 1 thread at a time (so by having one of these structures allocated by each
calling thread, you can have several tree searches going on in parallel). */

typedef struct NonRecursiveArgumentsStruct
{
  type_code keyType;
  AVLDupComparisonFunctionPointer keyComparisonFunctionPntr;
  AVLDupThingRecord userKey1;
  AVLDupThingRecord userKey2;
  unsigned int userValue1;
  unsigned int userValue2;
  int howToCompareUserValue1;
  int howToCompareUserValue2;
  bool includeThingEqualToStart;
  bool includeThingEqualToEnd;
  AVLDupIterationCallbackFunctionPointer iterationCallback;
  void *extraUserData;
  void * (*mallocPntr) (const void *pool_id, size_t size);
  void (*freePntr) (const void *pool_id, void *ptr);
  const void *poolID;
} NonRecursiveArgumentsRecord, *NonRecursiveArgumentsPointer;



/******************************************************************************
 * Copy an array of things contents from one thing to another, allocating
 * memory as needed (the caller provides both thing arrays).  If memory runs
 * out while copying an array of things, the partial allocations will be
 * undone, this function will return FALSE and the contents of the destination
 * will be garbage.  Assumes that the contents of the destination on entry are
 * uninitialised (so free any strings in the destination yourself if you were
 * using it to hold strings before calling this function).  Doesn't work if
 * source and destination are the same Thing.  Allows NULL allocation and free
 * routines, which means that anything which requires allocation will fail.
 */

bool AVLDupCopyThingArray (
  AVLDupThingPointer DestThingPntr,
  AVLDupThingConstPointer SourceThingPntr,
  type_code ThingType,
  int NumberOfThings,
  void * (*MallocFunctionPntr) (const void *pool_id, size_t size),
  void (*FreeFunctionPntr) (const void *pool_id, void *ptr),
  const void *PoolID)
{
  int         i;
  const char *SourceStringPntr;
  int         StringLength;
  char       *TempStringPntr;

  if (DestThingPntr == NULL || SourceThingPntr == NULL || NumberOfThings <= 0)
    return true; /* Nothing to do. */

  /* Most simple things can be copied directly. */

  if (ThingType != B_STRING_TYPE && ThingType != B_MIME_STRING_TYPE)
  {
    while (NumberOfThings-- > 0)
      *DestThingPntr++ = *SourceThingPntr++;
    return true;
  }

  /* Strings need extra storage allocated for long strings, but just get
  copied for short strings, and NULL strings get converted to zero length
  strings and stored as short strings. */

  TempStringPntr = (char *) 42; /* Initialise to non-NULL, NULL means error. */

  for (i = 0; i < NumberOfThings; i++, DestThingPntr++, SourceThingPntr++)
  {
    SourceStringPntr = AVLDupGetStringPntrFromThing (*SourceThingPntr);

    /* Clear out the Thing to entirely zero.  This makes it a short string,
    containing a zero length string. */

    DestThingPntr->int64Thing = 0;

    if (SourceStringPntr == NULL)
    {
      /*
      DestThingPntr->shortStringThing[0] = 0; / * Empty string. * /
      DestThingPntr->shortStringThing[7] = 0; / * Marks it as short string. * /
      */
    }
    else
    {
      StringLength = strlen (SourceStringPntr);

      if (StringLength <= 7)
      {
        /* Copy a short string or convert a long string to a short string. */

        strcpy (DestThingPntr->shortStringThing, SourceStringPntr);
        /* DestThingPntr->shortStringThing[7] = 0; Marks it as short string. */
      }
      else /* Long string, need to allocate memory. */
      {
        if (MallocFunctionPntr == NULL)
          TempStringPntr = NULL;
        else
          TempStringPntr = MallocFunctionPntr (PoolID, StringLength + 1);
        if (TempStringPntr == NULL)
          break; /* Oops, out of memory, abort to cleanup! */
        strcpy (TempStringPntr, SourceStringPntr);
        DestThingPntr->longStringThing.stringPntr = TempStringPntr;
        DestThingPntr->longStringThing.isLongString = true;
      }
    }
  }

  /* If we ran out of memory, undo all the allocations for all the previous
  long strings in the array. */

  if (TempStringPntr != NULL)
    return true; /* Successfully did all string copies. */

  for (i--, DestThingPntr--; i >= 0; i--, DestThingPntr--)
  {
    if (DestThingPntr->longStringThing.isLongString)
      FreeFunctionPntr (PoolID, DestThingPntr->longStringThing.stringPntr);
  }

  return false;
}



/******************************************************************************
 * This utility function just goes through the given number of array elements
 * and frees the long strings, does nothing for most other data types, the
 * result is an array of zeroes.  Throws away the memory if the memory free
 * function is NULL.
 */

void AVLDupFreeThingArray (
  AVLDupThingPointer ThingPntr,
  type_code          ThingType,
  int                NumberOfThings,
  void             (*FreeFunctionPntr) (const void *pool_id, void *ptr),
  const void        *PoolID)
{
  AVLDupThingPointer FreePntr;
  uint32             i;

  if (ThingPntr == NULL || NumberOfThings <= 0)
    return; /* Nothing to do. */

  if (FreeFunctionPntr != NULL &&
  (ThingType == B_STRING_TYPE || ThingType == B_MIME_STRING_TYPE))
  {
    for (i = NumberOfThings, FreePntr = ThingPntr;
    i > 0;
    i--, FreePntr++)
    {
      if (FreePntr->longStringThing.isLongString)
        FreeFunctionPntr (PoolID, FreePntr->longStringThing.stringPntr);
    }
  }

  if (NumberOfThings == 1)
    ThingPntr->int64Thing = 0;
  else
    memset (ThingPntr, 0, NumberOfThings * sizeof (AVLDupThingRecord));
}



/******************************************************************************
 * Internal function for converting a string to a date.  At the moment it just
 * understands one simple format, an integer describing the number of seconds
 * since 00:00:00 UTC, January 1, 1970.  I would like to use parsedate (from
 * INN's library, see http://www.isc.org/products/INN/ for the latest version
 * and use Bison to convert parsedate.y into parsedate.c) but parsedate
 * requires localtime () and all that time zone machinery which isn't available
 * in kernel land.  You can only get the time in seconds since the start of the
 * standard epoch in 1970 (time_t is an int32 which can be negative, which
 * takes you back to 1902, or positive, which goes up to 2038).  You can find
 * out the time zone with an undocumented function which I think has the
 * prototype: status_t get_tzfilename (buffer, buffersize), which returns a
 * path to a tzfile style file describing the time zone.  But processing that
 * means essentially implementing localtime().  So, instead we will just use
 * one simple format for dates.  If people want others, then the date parsing
 * can be enhanced later.  If parsing fails, returns a date of zero (so things
 * with incorrect dates appear in the middle of the sort order, in 1970).
 */

static time_t ConvertStringToTimeType (const char *StringPntr)
{
  char  *EndPntr;
  time_t SecondsSinceEpoch; /* Really an int32 */

  SecondsSinceEpoch = strtol (StringPntr, &EndPntr, 10);
  return SecondsSinceEpoch;
}


static bool ConvertTimeTypeToString (
  time_t SecondsSinceEpoch,
  char *StringPntr,
  int StringSize)
{
  if (StringSize < 80)
    return false;

  sprintf (StringPntr,
    "%ld seconds since January 1, 1970, Coordinated Universal Time.",
    SecondsSinceEpoch);

  return true;
}



/******************************************************************************
 * A utility function for converting a Human readable string to a Thing.
 * Returns TRUE if successful, FALSE if it ran out of memory or something else
 * went wrong.  Really should be international strings, but that's difficult to
 * do in kernel land.
 */

bool AVLDupConvertStringToThing (
  const char *StringPntr,
  AVLDupThingPointer ThingPntr,
  type_code ThingType,
  void * (*MallocFunctionPntr) (const void *pool_id, size_t size),
  void (*FreeFunctionPntr) (const void *pool_id, void *ptr),
  const void *PoolID)
{
  char             *EndPntr = NULL;
  int32             TempInt32;
  AVLDupThingRecord TempThing;
  uint32            TempUint32;

  if (ThingPntr == NULL)
    return false;

  ThingPntr->int64Thing = 0; /* Set unused bits to zero. */

  switch (ThingType) /* Listed in mostly alphabetical order. */
  {
    case B_BOOL_TYPE:
      if (strcasecmp (StringPntr, "Yes") == 0 ||
      strcasecmp (StringPntr, "Oui") == 0 ||
      strcasecmp (StringPntr, "True") == 0 ||
      strcasecmp (StringPntr, "T") == 0 ||
      strcasecmp (StringPntr, "Y") == 0)
        ThingPntr->uint8Thing = true;
      else if (strcasecmp (StringPntr, "No") == 0 ||
      strcasecmp (StringPntr, "Non") == 0 ||
      strcasecmp (StringPntr, "False") == 0 ||
      strcasecmp (StringPntr, "F") == 0 ||
      strcasecmp (StringPntr, "N") == 0)
        ThingPntr->uint8Thing = false;
      else
        ThingPntr->uint8Thing = (0 != strtol (StringPntr, &EndPntr, 10));
      return true;

    case B_CHAR_TYPE:
      ThingPntr->uint8Thing = StringPntr[0];
      return true;

    case B_DOUBLE_TYPE:
      ThingPntr->doubleThing = strtod (StringPntr, &EndPntr);
      return true;

    case B_FLOAT_TYPE:
      ThingPntr->floatThing = strtod (StringPntr, &EndPntr);
      return true;

    case B_INT8_TYPE:
      TempInt32 = strtol (StringPntr, &EndPntr, 10);
      if (TempInt32 < -128)
        ThingPntr->int8Thing = -128;
      else if (TempInt32 > 127)
        ThingPntr->int8Thing = 127;
      else
        ThingPntr->int8Thing = TempInt32;
      return true;

    case B_INT16_TYPE:
      TempInt32 = strtol (StringPntr, &EndPntr, 10);
      if (TempInt32 < -32768)
        ThingPntr->int16Thing = -32768;
      else if (TempInt32 > 32767)
        ThingPntr->int16Thing = 32767;
      else
        ThingPntr->int16Thing = TempInt32;
      return true;

    case B_INT32_TYPE:
    case B_SSIZE_T_TYPE:
      ThingPntr->int32Thing = strtol (StringPntr, &EndPntr, 10);
      return true;

    case B_INT64_TYPE:
    case B_OFF_T_TYPE:
      ThingPntr->int64Thing = strtoll (StringPntr, &EndPntr, 10);
      return true;

    case B_STRING_TYPE:
    case B_MIME_STRING_TYPE:
      TempThing.longStringThing.stringPntr = (char *) StringPntr;
      TempThing.longStringThing.isLongString = true;
      return AVLDupCopyThingArray (ThingPntr, &TempThing, ThingType, 1,
        MallocFunctionPntr, FreeFunctionPntr, PoolID);

    case B_TIME_TYPE:
      ThingPntr->timeThing = ConvertStringToTimeType (StringPntr);
      return true;

    case B_UINT8_TYPE:
      TempUint32 = strtoul (StringPntr, &EndPntr, 10);
      if (TempUint32 > 255)
        ThingPntr->uint8Thing = 255;
      else
        ThingPntr->uint8Thing = TempUint32;
      return true;

    case B_UINT16_TYPE:
      TempUint32 = strtoul (StringPntr, &EndPntr, 10);
      if (TempUint32 > 65535)
        ThingPntr->uint16Thing = 65535;
      else
        ThingPntr->uint16Thing = TempUint32;
      return true;

    case B_UINT32_TYPE:
    case B_SIZE_T_TYPE:
      ThingPntr->uint32Thing = strtoul (StringPntr, &EndPntr, 10);
      return true;

    case B_UINT64_TYPE:
      ThingPntr->uint64Thing = strtoull (StringPntr, &EndPntr, 10);
      return true;
  }

  return false;
}



/******************************************************************************
 * Convert a Thing into a Human readable string.  StringSize is the size of
 * your buffer, which will receive the NUL terminated string.  Returns TRUE if
 * successful, FALSE if your buffer is too small or something else is wrong.
 */

bool AVLDupConvertThingToString (
  AVLDupThingConstPointer ThingPntr,
  type_code ThingType,
  char *StringPntr,
  int StringSize)
{
  const char *TempStringPntr;

  if (ThingPntr == NULL || StringPntr == NULL)
    return false;

  switch (ThingType)
  {
    case B_BOOL_TYPE:
      if (StringSize < 6) return false;
      if (ThingPntr->uint8Thing)
        strcpy (StringPntr, "True");
      else
        strcpy (StringPntr, "False");
      return true;

    case B_CHAR_TYPE:
      if (StringSize < 2) return false;
      StringPntr[0] = ThingPntr->uint8Thing;
      StringPntr[1] = 0;
      return true;

    case B_DOUBLE_TYPE:
      if (StringSize < 24) return false;
      sprintf (StringPntr, "%g", ThingPntr->doubleThing);
      return true;

    case B_FLOAT_TYPE:
      if (StringSize < 24) return false;
      sprintf (StringPntr, "%g", (double) ThingPntr->floatThing);
      return true;

    case B_INT8_TYPE:
      if (StringSize < 5) return false;
      sprintf (StringPntr, "%d", (int) ThingPntr->int8Thing);
      return true;

    case B_INT16_TYPE:
      if (StringSize < 7) return false;
      sprintf (StringPntr, "%d", (int) ThingPntr->int16Thing);
      return true;

    case B_INT32_TYPE:
    case B_SSIZE_T_TYPE:
      if (StringSize < 12) return false;
      sprintf (StringPntr, "%ld", ThingPntr->int32Thing);
      return true;

    case B_INT64_TYPE:
    case B_OFF_T_TYPE:
      if (StringSize < 24) return false;
      sprintf (StringPntr, "%Ld", ThingPntr->int64Thing);
      return true;

    case B_STRING_TYPE:
    case B_MIME_STRING_TYPE:
      if (StringSize < 4) return false;
      TempStringPntr = AVLDupGetStringPntrFromThing (*ThingPntr);
      if (TempStringPntr == NULL)
        *StringPntr = 0; /* Return an empty string instead of NULL. */
      else if (strlen (TempStringPntr) < StringSize)
        strcpy (StringPntr, TempStringPntr);
      else /* Destination string is too short. */
      {
        strncpy (StringPntr, TempStringPntr, StringSize);
        strcpy (StringPntr + (StringSize-4), "...");
      }
      return true;

    case B_TIME_TYPE:
      return ConvertTimeTypeToString (ThingPntr->timeThing,
        StringPntr, StringSize);

    case B_UINT8_TYPE:
      if (StringSize < 5) return false;
      sprintf (StringPntr, "%u", (unsigned int) ThingPntr->uint8Thing);
      return true;

    case B_UINT16_TYPE:
      if (StringSize < 7) return false;
      sprintf (StringPntr, "%u", (unsigned int) ThingPntr->uint16Thing);
      return true;

    case B_UINT32_TYPE:
    case B_SIZE_T_TYPE:
      if (StringSize < 12) return false;
      sprintf (StringPntr, "%lu", ThingPntr->uint32Thing);
      return true;

    case B_UINT64_TYPE:
      if (StringSize < 24) return false;
      sprintf (StringPntr, "%Lu", ThingPntr->uint64Thing);
      return true;
  }

  return false;
}



/******************************************************************************
 * Returns the number of bytes of storage for the given data type.  Returns -1
 * if it isn't a recognized type.  Returns 0 for variable length types
 * (strings).
 */

int AVLDupGetThingTypeSize (type_code ThingType)
{
  switch (ThingType)
  {
    case B_BOOL_TYPE:
    case B_CHAR_TYPE:
    case B_INT8_TYPE:
    case B_UINT8_TYPE:
      return 1;

    case B_INT16_TYPE:
    case B_UINT16_TYPE:
      return 2;

    case B_FLOAT_TYPE:
    case B_INT32_TYPE:
    case B_UINT32_TYPE:
    case B_SIZE_T_TYPE:
    case B_SSIZE_T_TYPE:
    case B_TIME_TYPE:
      return 4;

    case B_DOUBLE_TYPE:
    case B_INT64_TYPE:
    case B_UINT64_TYPE:
    case B_OFF_T_TYPE:
      return 8;

    case B_STRING_TYPE:
    case B_MIME_STRING_TYPE:
      return 0;
  }

  return -1;
}



/* A series of value comparison functions for internal use.  They essentially
return the sign of the operation A - B, so the result is >0 for A > B,
<0 for A < B and 0 for A equals B. */

static int CompareDouble (AVLDupThingConstPointer A, AVLDupThingConstPointer B)
{
  double Delta;

  Delta = A->doubleThing - B->doubleThing;
  if (Delta < 0)
    return -1;
  else if (Delta > 0)
    return 1;
  return 0;
}


static int CompareFloat (AVLDupThingConstPointer A, AVLDupThingConstPointer B)
{
  float Delta;

  Delta = A->floatThing - B->floatThing;
  if (Delta < 0)
    return -1;
  else if (Delta > 0)
    return 1;
  return 0;
}


static int CompareInt8 (AVLDupThingConstPointer A, AVLDupThingConstPointer B)
{
  if (A->int8Thing < B->int8Thing)
    return -1;
  else if (A->int8Thing > B->int8Thing)
    return 1;
  return 0;
}


static int CompareInt16 (AVLDupThingConstPointer A, AVLDupThingConstPointer B)
{
  if (A->int16Thing < B->int16Thing)
    return -1;
  else if (A->int16Thing > B->int16Thing)
    return 1;
  return 0;
}


static int CompareInt32 (AVLDupThingConstPointer A, AVLDupThingConstPointer B)
{
  if (A->int32Thing < B->int32Thing)
    return -1;
  else if (A->int32Thing > B->int32Thing)
    return 1;
  return 0;
}


static int CompareInt64 (AVLDupThingConstPointer A, AVLDupThingConstPointer B)
{
  if (A->int64Thing < B->int64Thing)
    return -1;
  else if (A->int64Thing > B->int64Thing)
    return 1;
  return 0;
}


static int CompareString (AVLDupThingConstPointer A, AVLDupThingConstPointer B)
{
  const char *StringAPntr;
  const char *StringBPntr;

  StringAPntr = AVLDupGetStringPntrFromThing (*A);
  StringBPntr = AVLDupGetStringPntrFromThing (*B);

  /* Handle NULL pointer - represents a string smaller than all others,
  though you shouldn't see it in practice (normally it gets converted to
  an empty string in short string format). */

  if (StringAPntr == NULL && StringBPntr == NULL)
    return 0;

  if (StringAPntr == NULL)
    return -1;

  if (StringBPntr == NULL)
    return 1;

  /* Use an ordinary strcmp for the rest of the comparison.  We are comparing
  UTF-8 strings, so the ordering of multibyte characters could be different
  than what strcmp returns, but we'll leave it as it is for now.  Also, we
  could implement a comparison which ignores case. */

  return strcmp (StringAPntr, StringBPntr);
}


static int CompareUint8 (AVLDupThingConstPointer A, AVLDupThingConstPointer B)
{
  if (A->uint8Thing < B->uint8Thing)
    return -1;
  else if (A->uint8Thing > B->uint8Thing)
    return 1;
  return 0;
}


static int CompareUint16 (AVLDupThingConstPointer A, AVLDupThingConstPointer B)
{
  if (A->uint16Thing < B->uint16Thing)
    return -1;
  else if (A->uint16Thing > B->uint16Thing)
    return 1;
  return 0;
}


static int CompareUint32 (AVLDupThingConstPointer A, AVLDupThingConstPointer B)
{
  if (A->uint32Thing < B->uint32Thing)
    return -1;
  else if (A->uint32Thing > B->uint32Thing)
    return 1;
  return 0;
}


static int CompareUint64 (AVLDupThingConstPointer A, AVLDupThingConstPointer B)
{
  if (A->uint64Thing < B->uint64Thing)
    return -1;
  else if (A->uint64Thing > B->uint64Thing)
    return 1;
  return 0;
}



/* Internal and external utility to find a comparison function which is
appropriate for the specified type of data.  Returns NULL if none exists. */

AVLDupComparisonFunctionPointer AVLDupGetComparisonFunctionForType (
  type_code KeyType)
{
  switch (KeyType)
  {
    case B_DOUBLE_TYPE:
      return CompareDouble;

    case B_FLOAT_TYPE:
      return CompareFloat;

    case B_INT8_TYPE:
      return CompareInt8;

    case B_INT16_TYPE:
      return CompareInt16;

    case B_INT32_TYPE:
    case B_SSIZE_T_TYPE:
    case B_TIME_TYPE:
      return CompareInt32;

    case B_INT64_TYPE:
    case B_OFF_T_TYPE:
      return CompareInt64;

    case B_STRING_TYPE:
    case B_MIME_STRING_TYPE:
      return CompareString;

    case B_UINT8_TYPE:
    case B_BOOL_TYPE:
    case B_CHAR_TYPE:
      return CompareUint8;

    case B_UINT16_TYPE:
      return CompareUint16;

    case B_UINT32_TYPE:
    case B_SIZE_T_TYPE:
      return CompareUint32;

    case B_UINT64_TYPE:
      return CompareUint64;
  }

  return NULL;
}



/******************************************************************************
 * Create a new empty AVLDupTree.  Returns the new initialised tree header or
 * NULL if out of memory or the key data types are unsupported.  Will use the
 * specified memory pool and associated access functions to allocate and free
 * memory for both the tree nodes, and string based keys.
 */

AVLDupTreePointer AVLDupAllocTree (
  type_code KeyType,
  void * (*MallocFunctionPntr) (const void *pool_id, size_t size),
  void (*FreeFunctionPntr) (const void *pool_id, void *ptr),
  const void *PoolID)
{
  AVLDupTreePointer  NewTree;

  NewTree = MallocFunctionPntr (PoolID, sizeof (AVLDupTreeRecord));
  if (NewTree == NULL) goto ErrorExit;

  /* Set up some initial values so that cleanup of
  further errors won't cause a crash. */

  NewTree->rootPntr = NULL;
  NewTree->count = 0;
  NewTree->mallocPntr = MallocFunctionPntr;
  NewTree->freePntr = FreeFunctionPntr;
  NewTree->poolID = PoolID;

  /* Set up the data types and corresponding comparison functions. */

  NewTree->keyType = KeyType;
  NewTree->keyComparisonFunctionPntr =
    AVLDupGetComparisonFunctionForType (KeyType);
  if (NewTree->keyComparisonFunctionPntr == NULL) goto ErrorExit;

  return NewTree;


ErrorExit: /* Deallocate partial allocations and return NULL. */
  AVLDupFreeTree (NewTree);
  return NULL;
}



/* Internal function for removing nodes.  Deallocates the node's children, the
node's key, and the node itself.  It needs the tree only for the type of the
key and the memory freeing function.  It does not update the tree record at
all, nor rebalance the tree, nor fix up parent nodes of the deleted nodes. */

static void AVLDupRecursiveDeallocateNodes (
  AVLDupTreePointer  TreePntr,
  AVLDupNodePointer  CurrentNode)
{
  if (CurrentNode == NULL)
    return;

  /* Delete child nodes. */

  AVLDupRecursiveDeallocateNodes (TreePntr, CurrentNode->smallerChildPntr);
  AVLDupRecursiveDeallocateNodes (TreePntr, CurrentNode->largerChildPntr);

  /* Deallocate associated data. */

  AVLDupFreeThingArray (&CurrentNode->key, TreePntr->keyType, 1,
    TreePntr->freePntr, TreePntr->poolID);

  /* Finally get rid of the node itself. */

  memset (CurrentNode, 0, sizeof (AVLDupNodeRecord));
  TreePntr->freePntr (TreePntr->poolID, CurrentNode);
}



/******************************************************************************
 * Deallocate an AVLDupTree.  Deallocates all the nodes and associated strings
 * if any, then frees the tree header record.  Safe to pass in NULL.
 */

void AVLDupFreeTree (AVLDupTreePointer TreePntr)
{
  void (*FreeFunctionPntr) (const void *pool_id, void *ptr);
  const void *PoolID;

  if (TreePntr != NULL)
  {
    if (TreePntr->rootPntr != NULL)
      AVLDupRecursiveDeallocateNodes (TreePntr, TreePntr->rootPntr);

    FreeFunctionPntr = TreePntr->freePntr;
    PoolID = TreePntr->poolID;
    memset (TreePntr, 0, sizeof (AVLDupTreeRecord));
    FreeFunctionPntr (PoolID, TreePntr);
  }
}



/******************************************************************************
 * Returns the number of key/value pairs in the tree.
 */

unsigned int AVLDupGetTreeCount (AVLDupTreePointer TreePntr)
{
  if (TreePntr != NULL)
    return TreePntr->count;

  return 0;
}



/* Internal function for setting the height of a node to be the larger of
the child nodes' heights, plus one. */

static void AVLDupRecalculateNodeHeight (AVLDupNodePointer CurrentNode)
{
  unsigned int LeftHeight;
  unsigned int RightHeight;

  if (CurrentNode->smallerChildPntr == NULL)
    LeftHeight = 0;
  else
    LeftHeight = CurrentNode->smallerChildPntr->height;

  if (CurrentNode->largerChildPntr == NULL)
    RightHeight = 0;
  else
    RightHeight = CurrentNode->largerChildPntr->height;

  if (LeftHeight < RightHeight)
    CurrentNode->height = RightHeight + 1;
  else
    CurrentNode->height = LeftHeight + 1;
}



/* Internal function for making the left child into the root of the (sub)tree,
pushing the former root down to the right.  The Dr. Dobbs article calls this a
Right Rotation.  It's useful for balancing when the left subtree is deeper than
the right subtree.  It maintains the binary tree property (left is less, right
is greater) but doesn't maintain the AVL properties and may change the height
of the (sub)tree (the parent will need to recalculate its height; however the
heights of the nodes involved will be updated by this function).

The shuffle works like this: the left child becomes the root of the tree, and
since the old root was larger than the left child, it can be safely added as
the new right child of the new root.  So, what happens to the old right child
of the new root?  We know it is less than the old root, so we add it as a left
child of the old root (conveniently free because the left child was moved up
and become the new root).  The left child of the old left child is still
attached where it was, so it is now the left child of the new root.  The same
goes for the right child of the old root, which remains attached to that node
and is thus now the right child of the new right child.

Here's an ASCII picture, 1 and 2 are nodes (1 has a smaller key than 2), and
A, B, C are subtrees (which can be NULL):

  Before        After

     2           1
    / \         / \
   1   C       A   2
  / \             / \
 A   B           B   C

*/

static void AVLDupRaiseLeftChild (AVLDupNodePointer *ParentNodePntrPntr)
{
  AVLDupNodePointer  Node1Pntr;
  AVLDupNodePointer  Node2Pntr;

  Node2Pntr = *ParentNodePntrPntr;
  if (Node2Pntr == NULL) return; /* Nothing to do. */

  Node1Pntr = Node2Pntr->smallerChildPntr;
  if (Node1Pntr == NULL) return; /* Oops, can't raise "nothing" up. */

  /* Do the BST rotation. */

  *ParentNodePntrPntr = Node1Pntr;
  Node2Pntr->smallerChildPntr = Node1Pntr->largerChildPntr;
  Node1Pntr->largerChildPntr = Node2Pntr;

  /* Fix up the heights of the nodes which have been changed, except for the
  parent node, which the caller should fix up.  Do it in order of increasing
  tree hierarchy, from lower nodes to higher nodes so that it gets the right
  value! */

  AVLDupRecalculateNodeHeight (Node2Pntr);
  AVLDupRecalculateNodeHeight (Node1Pntr);
}



/* Internal function for making the right child into the root of the (sub)tree,
pushing the former root down to the left.

Here's an ASCII picture, 1 and 2 are nodes (1 has a smaller key than 2), and
A, B, C are subtrees (which can be NULL):

 Before        After

   1              2
  / \            / \
 A   2          1   C
    / \        / \
   B   C      A   B

*/

static void AVLDupRaiseRightChild (AVLDupNodePointer *ParentNodePntrPntr)
{
  AVLDupNodePointer  Node1Pntr;
  AVLDupNodePointer  Node2Pntr;

  Node1Pntr = *ParentNodePntrPntr;
  if (Node1Pntr == NULL) return; /* Nothing to do. */

  Node2Pntr = Node1Pntr->largerChildPntr;
  if (Node2Pntr == NULL) return; /* Oops, can't raise "nothing" up. */

  /* Do the BST rotation. */

  *ParentNodePntrPntr = Node2Pntr;
  Node1Pntr->largerChildPntr = Node2Pntr->smallerChildPntr;
  Node2Pntr->smallerChildPntr = Node1Pntr;

  /* Fix up the heights of the nodes which have been changed. */

  AVLDupRecalculateNodeHeight (Node1Pntr);
  AVLDupRecalculateNodeHeight (Node2Pntr);
}



/* Internal function for rebalancing the left and right subtrees if their
heights differ by more than 1 so that they end up differing by at most 1. It
also updates the height of the node in all cases.  This function is called
after an addition or deletion is made to the tree, as part of the
addition/deletion recursive chain so that it rebalances the changed parts all
the way from the added/deleted node up to the root. */

static void AVLDupFixupSubtrees (AVLDupNodePointer *ParentNodePntrPntr)
{
  AVLDupNodePointer CurrentNode;
  int               Delta;
  unsigned int      LeftHeight;
  AVLDupNodePointer RaiseNode;
  unsigned int      RightHeight;

  CurrentNode = *ParentNodePntrPntr;
  if (CurrentNode == NULL) return; /* Nothing to do. */

  if (CurrentNode->smallerChildPntr == NULL)
    LeftHeight = 0;
  else
    LeftHeight = CurrentNode->smallerChildPntr->height;

  if (CurrentNode->largerChildPntr == NULL)
    RightHeight = 0;
  else
    RightHeight = CurrentNode->largerChildPntr->height;

  Delta = (int) LeftHeight - (int) RightHeight;

  if (Delta <= -2)
  {
    /* Left side is deficient in height, raise the right side node up to be the
    new root of the subtree, reducing the height of the right side of the tree
    relative to the left side, and increasing the left side height by pushing
    down the old root of the subtree into the left side. */

    /* But before raising the right child, make sure the right side's right
    child (a grandchild) is the deeper of the two grandchildren otherwise the
    raised node will have a right child which isn't deep enough (the raised
    right node keeps its right grandchild) and the AVL property won't be true
    (right grandchild will be off by more than 1 in height from the left
    grandchild). */

    RaiseNode = CurrentNode->largerChildPntr;
    LeftHeight = (RaiseNode->smallerChildPntr == NULL) ?
      0 : RaiseNode->smallerChildPntr->height;
    RightHeight = (RaiseNode->largerChildPntr == NULL) ?
      0 : RaiseNode->largerChildPntr->height;
    if (RightHeight < LeftHeight)
      AVLDupRaiseLeftChild (&CurrentNode->largerChildPntr);

    AVLDupRaiseRightChild (ParentNodePntrPntr);
  }
  else if (Delta >= 2)
  {
    /* Right side is deficient in height, lower right / raise left side. */

    RaiseNode = CurrentNode->smallerChildPntr;
    LeftHeight = (RaiseNode->smallerChildPntr == NULL) ?
      0 : RaiseNode->smallerChildPntr->height;
    RightHeight = (RaiseNode->largerChildPntr == NULL) ?
      0 : RaiseNode->largerChildPntr->height;
    if (LeftHeight < RightHeight) /* Avoid AVL grandchild problem. */
      AVLDupRaiseRightChild (&CurrentNode->smallerChildPntr);

    AVLDupRaiseLeftChild (ParentNodePntrPntr);
  }
  else /* Delta is -1, 0, or +1.  No balancing needed, just update height. */
  {
    /* Rather than calling AVLDupRecalculateNodeHeight, we can do it faster
    right here.  Use the larger of the child heights + 1. */

    if (Delta < 0)
      CurrentNode->height = RightHeight + 1;
    else
      CurrentNode->height = LeftHeight + 1;
  }
}



/* Internal function for adding a new node.  It recursively traverses the tree
until the key/value is found (in which case it does nothing if it is exactly
the same as an existing node), or it finds an empty spot where it can add the
new node.  The parent node's pointer to the child being considered is passed in
so that we can change the parent's child pointer when the tree gets rearranged
as part of the balancing, without knowing if the parent is a node (left or
right subree) or the tree header.  Returns AVLADD_ADDED_A_NODE if it added a
node, AVLADD_ALREADY_IN_TREE if it found a duplicate key/value (does nothing to
tree), and AVLADD_OUT_OF_MEMORY if it ran out of memory (also does nothing to
the existing tree). */

static AVLDupAddReturnCode AVLDupRecursiveAddNode (
  NonRecursiveArgumentsPointer ArgsPntr,
  AVLDupNodePointer           *ParentsChildPntrPntr)
{
  int                 ComparisonResult;
  AVLDupNodePointer   CurrentNode;
  AVLDupAddReturnCode ErrorCode;
  AVLDupNodePointer   NewNode;

  CurrentNode = *ParentsChildPntrPntr;
  if (CurrentNode == NULL) /* Empty tree or new leaf situation. */
  {
    /* Create a new node and add it to the tree. */

    NewNode =
      ArgsPntr->mallocPntr (ArgsPntr->poolID, sizeof (AVLDupNodeRecord));
    if (NewNode == NULL)
      return AVLADD_OUT_OF_MEMORY;

    /* Copy the key and value to the new node. */

    if (!AVLDupCopyThingArray (&NewNode->key, &ArgsPntr->userKey1,
      ArgsPntr->keyType, 1, ArgsPntr->mallocPntr, ArgsPntr->freePntr,
      ArgsPntr->poolID))
    {
      ArgsPntr->freePntr (ArgsPntr->poolID, NewNode);
      return AVLADD_OUT_OF_MEMORY;
    }
    NewNode->value = ArgsPntr->userValue1;

    /* Set up the remaining fields in the new node. */

    NewNode->smallerChildPntr = NULL;
    NewNode->largerChildPntr = NULL;
    NewNode->height = 1;

    *ParentsChildPntrPntr = NewNode;
    return AVLADD_ADDED_A_NODE;
  }

  /* Adding a node to an existing CurrentNode.  See if the key/value pair is
  less than, equal to, or greater than the CurrentNode's key/value, which will
  decide which subtree to add the new node to. */

  ComparisonResult = ArgsPntr->keyComparisonFunctionPntr (
    &ArgsPntr->userKey1, &CurrentNode->key);

  if (ComparisonResult == 0) /* Equal keys, use the value to decide. */
    ComparisonResult = ((int) ArgsPntr->userValue1) - (int) CurrentNode->value;

  /* Key/value pair is totally equal to the current node.  Just do nothing. */

  if (ComparisonResult == 0)
    return AVLADD_ALREADY_IN_TREE;

  /* Add the new key/value to one of the subtrees. */

  if (ComparisonResult < 0)
  {
    /* Key is smaller than the current node, add the new node to the left
    subtree of the current node. */

    ErrorCode =
      AVLDupRecursiveAddNode (ArgsPntr, &CurrentNode->smallerChildPntr);
  }
  else
  {
    /* Key is larger than the current node, add the new node to the right
    subtree of the current node. */

    ErrorCode =
      AVLDupRecursiveAddNode (ArgsPntr, &CurrentNode->largerChildPntr);
  }

  /* If the node wasn't added, the tree hasn't changed. */

  if (ErrorCode != AVLADD_ADDED_A_NODE)
    return ErrorCode;

  /* A new node was added as a child.  Do post-addition fixups.  Recompute
  heights and rebalance the tree if needed. */

  AVLDupFixupSubtrees (ParentsChildPntrPntr);

  return AVLADD_ADDED_A_NODE;
}



/******************************************************************************
 * Adds a key/value pair to the AVLDupTree.  Returns AVLADD_ADDED_A_NODE or
 * AVLADD_ALREADY_IN_TREE if successful (zero or positive numbers),
 * AVLADD_OUT_OF_MEMORY (a negative number) if it ran out of memory.  When it
 * returns AVLADD_ALREADY_IN_TREE it means you specified an existing key/value
 * pair, and thus nothing was done to the tree.  The type of the key is assumed
 * to be the same as defined in the tree header record.  The key is copied so
 * that you can deallocate it after calling this function.  If you specified a
 * string thing, the string will be copied by the tree routines, so you can
 * free your string too.  It will also convert internally between short and
 * long strings, so you can always pass in a long string if you wish.
 */

AVLDupAddReturnCode AVLDupAdd (
  AVLDupTreePointer  TreePntr,
  AVLDupThingPointer Key,
  unsigned int       Value)
{
  NonRecursiveArgumentsRecord Arguments;
  AVLDupAddReturnCode         ReturnCode;

  if (TreePntr == NULL || Key == NULL)
    return AVLADD_BAD_ARGUMENTS;

  Arguments.keyType = TreePntr->keyType;
  Arguments.keyComparisonFunctionPntr = TreePntr->keyComparisonFunctionPntr;
  Arguments.userKey1 = *Key;
  Arguments.userValue1 = Value;
  Arguments.mallocPntr = TreePntr->mallocPntr;
  Arguments.freePntr = TreePntr->freePntr;
  Arguments.poolID = TreePntr->poolID;

  ReturnCode = AVLDupRecursiveAddNode (&Arguments, &TreePntr->rootPntr);

  if (ReturnCode == AVLADD_ADDED_A_NODE)
    TreePntr->count++;

  return ReturnCode;
}



/* An internal recursive function for finding the successor to some distant
parent node (just follow the left links until NULL is encountered), and remove
it from the tree (don't deallocate, just unlink and fix up its children).
Returns the removed node. */

static AVLDupNodePointer AVLDupRecursiveRipOutSuccessor (
  AVLDupNodePointer *ParentsChildPntrPntr)
{
  AVLDupNodePointer CurrentNode;
  AVLDupNodePointer RippedOutNode;

  CurrentNode = *ParentsChildPntrPntr;

  if (CurrentNode->smallerChildPntr == NULL)
  {
    /* Unlink the node from the tree.  Don't need to fix up its child since we
    didn't change it.  But the caller will need to fix up its heights and
    rebalance too, as we changed a child. */

    *ParentsChildPntrPntr = CurrentNode->largerChildPntr;
    CurrentNode->largerChildPntr = NULL;
    return CurrentNode;
  }

  /* Not at the successor yet, continue on down. */

  RippedOutNode =
    AVLDupRecursiveRipOutSuccessor (&CurrentNode->smallerChildPntr);

  /* Since child trees may have been modified, need to recalculate our height
  and do rebalancing. */

  AVLDupFixupSubtrees (ParentsChildPntrPntr);

  return RippedOutNode;
}



/* An internal recursive function for finding and deleting a node matching the
given key/value pair.  Once found, if the node has zero or one children, it is
removed from the tree normally.  If it has two children, then it gets replaced
by the next larger key in the tree, which is removed from its old position in
the tree.  The next larger value is guaranteed to be a leaf node in the tree,
found by taking the right child, then all the left children until there are no
more left children.  That next-larger-value node is ripped out of the tree
(replaced by it's right child if any).  Then the rip-out traversal returns back
to the deletion point (fixing up the tree heights along the way), and the
replacement of the deleted node with the next larger node is made.  The
deletion function then returns, continuing to fix up heights along the return
chain.  The parent node's pointer to the node being considered is passed in so
that we can change the parent's child pointer when the tree gets rearranged as
part of the balancing, without knowing if the parent is a node (left or right
subree) or the tree header.  Returns TRUE if it deleted the node, FALSE if it
couldn't find it. */

static bool AVLDupRecursiveDeleteNodeFindIt (
  NonRecursiveArgumentsPointer ArgsPntr,
  AVLDupNodePointer *ParentsChildPntrPntr)
{
  int               ComparisonResult;
  AVLDupNodePointer CurrentNode;
  bool              ReturnCode;
  AVLDupNodePointer SuccessorNode;

  CurrentNode = *ParentsChildPntrPntr;
  if (CurrentNode == NULL) return false; /* Failed to find node to delete. */

  ComparisonResult = ArgsPntr->keyComparisonFunctionPntr (
    &ArgsPntr->userKey1, &CurrentNode->key);

  if (ComparisonResult == 0) /* Equal keys, use the value to decide. */
    ComparisonResult = ((int) ArgsPntr->userValue1) - (int) CurrentNode->value;

  if (ComparisonResult < 0)
    ReturnCode =
    AVLDupRecursiveDeleteNodeFindIt (ArgsPntr, &CurrentNode->smallerChildPntr);
  else if (ComparisonResult > 0)
    ReturnCode =
    AVLDupRecursiveDeleteNodeFindIt (ArgsPntr, &CurrentNode->largerChildPntr);
  else /* Found the key/value pair. */
  {
    if (CurrentNode->largerChildPntr == NULL)
    {
      /* Can delete this node directly, replacing it with the left subtree. */

      *ParentsChildPntrPntr = CurrentNode->smallerChildPntr;
    }
    else if (CurrentNode->smallerChildPntr == NULL)
    {
      /* Safe to replace the node with the right subtree. */

      *ParentsChildPntrPntr = CurrentNode->largerChildPntr;
    }
    else /* Has 2 children.  Replace with successor node. */
    {
      SuccessorNode = AVLDupRecursiveRipOutSuccessor (
        &CurrentNode->largerChildPntr);

      SuccessorNode->smallerChildPntr = CurrentNode->smallerChildPntr;
      SuccessorNode->largerChildPntr = CurrentNode->largerChildPntr;
      *ParentsChildPntrPntr = SuccessorNode;
    }

    /* Deallocate the old node, which matched our key/value pair. */

    AVLDupFreeThingArray (&CurrentNode->key, ArgsPntr->keyType, 1,
      ArgsPntr->freePntr, ArgsPntr->poolID);
    memset (CurrentNode, 0, sizeof (AVLDupNodeRecord));
    ArgsPntr->freePntr (ArgsPntr->poolID, CurrentNode);
    ReturnCode = true;
  }

  /* Recompute heights and rebalance the tree if needed. */

  AVLDupFixupSubtrees (ParentsChildPntrPntr);

  return ReturnCode;
}



/******************************************************************************
 * Deletes the given key/value pair.  Yes, you need to specify a value since
 * duplicate keys can't otherwise be distinguished.  Returns FALSE if it can't
 * find the key/value pair, TRUE if it deleted it.
 */

bool AVLDupDelete (
  AVLDupTreePointer TreePntr,
  AVLDupThingPointer Key,
  unsigned int Value)
{
  NonRecursiveArgumentsRecord Arguments;
  bool                        Successful;

  if (TreePntr == NULL || Key == NULL)
    return false;

  Arguments.keyType = TreePntr->keyType;
  Arguments.keyComparisonFunctionPntr = TreePntr->keyComparisonFunctionPntr;
  Arguments.userKey1 = *Key;
  Arguments.userValue1 = Value;
  Arguments.mallocPntr = TreePntr->mallocPntr;
  Arguments.freePntr = TreePntr->freePntr;
  Arguments.poolID = TreePntr->poolID;

  Successful =
    AVLDupRecursiveDeleteNodeFindIt (&Arguments, &TreePntr->rootPntr);

  if (Successful)
    TreePntr->count--;

  return Successful;
}



/* Internal recursive function for iterating over a subtree without doing any
tests.  Even assumes that the entry node is non-NULL.  Returns TRUE if
successful, FALSE if the user callback aborted the iteration. */

static bool AVLDupRecursiveSimpleIterate (
  NonRecursiveArgumentsPointer ArgsPntr,
  AVLDupNodePointer CurrentNode)
{
  /* Output left subtree first, since we are doing it in ascending order. */

  if (CurrentNode->smallerChildPntr != NULL)
  {
    if (!AVLDupRecursiveSimpleIterate (ArgsPntr,
    CurrentNode->smallerChildPntr))
      return false; /* User aborted. */
  }

  /* Output the middle node. */

  if (!ArgsPntr->iterationCallback (&CurrentNode->key, &CurrentNode->value,
  ArgsPntr->extraUserData))
    return false; /* The user requested an early abort of the iteration. */

  /* Finally the right subtree. */

  if (CurrentNode->largerChildPntr != NULL)
  {
    if (!AVLDupRecursiveSimpleIterate (ArgsPntr,
    CurrentNode->largerChildPntr))
      return false; /* User aborted. */
  }

  return true;
}



/* Recursively iterate over the tree in ascending order, cutting off traversals
which don't fit in the given range of keys.  Returns TRUE if it got to the end
of the range.  If TestLowerBound is TRUE then tests will be done against
userKey1 (the lower bound key).  If it is FALSE then we will assume that the
node and all its children are greater than or equal to userKey1 and avoid the
test.  Similarly if TestUpperBound is FALSE then we assume everything is less
than or equal to the upper bound.  If both are false, we revert to a simple
tree traversal and do no tests. */

static bool AVLDupRecursiveRangeIterate (
  NonRecursiveArgumentsPointer ArgsPntr,
  AVLDupNodePointer CurrentNode,
  bool TestLowerBound,
  bool TestUpperBound)
{
  int ComparisonLower;
  int ComparisonUpper;

  if (CurrentNode == NULL)
    return true; /* Successfully finished iterating the NULL tree. */

  if (!(TestLowerBound || TestUpperBound))
    return AVLDupRecursiveSimpleIterate (ArgsPntr, CurrentNode);

  /* For convenience in understanding the code, both high and low comparisons
  are done as (BoundsLimit - CurrentNodeKey).  Meaning ComparisonResult is
  less than zero for key bigger than the bounds limit, and so on. */

  if (TestLowerBound)
  {
    ComparisonLower = ArgsPntr->keyComparisonFunctionPntr (
    &ArgsPntr->userKey1, &CurrentNode->key);

    if (ComparisonLower == 0) /* Equal keys, use the value to decide. */
    {
      if (ArgsPntr->howToCompareUserValue1 == 0)
      {
        if (ArgsPntr->userValue1 < CurrentNode->value)
          ComparisonLower = -1;
        else if (ArgsPntr->userValue1 == CurrentNode->value)
          ComparisonLower = 0;
        else
          ComparisonLower = 1;
      }
      else /* Lower bound is plus or minus infinity as requested. */
        ComparisonLower = ArgsPntr->howToCompareUserValue1;
    }
  }
  else /* No comparison, current is always larger. */
    ComparisonLower = -1;

  if (TestUpperBound)
  {
    ComparisonUpper = ArgsPntr->keyComparisonFunctionPntr (
    &ArgsPntr->userKey2, &CurrentNode->key);

    if (ComparisonUpper == 0) /* Equal keys, use the value to decide. */
    {
      if (ArgsPntr->howToCompareUserValue2 == 0)
      {
        if (ArgsPntr->userValue2 < CurrentNode->value)
          ComparisonUpper = -1;
        else if (ArgsPntr->userValue2 == CurrentNode->value)
          ComparisonUpper = 0;
        else
          ComparisonUpper = 1;
      }
      else /* Upper bound is plus or minus infinity as requested. */
        ComparisonUpper = ArgsPntr->howToCompareUserValue2;
    }
  }
  else /* No comparison, current is always smaller. */
    ComparisonUpper = 1;

  /* Examine the left subtree.  The lower limit of the range has to be less
  than the current node's key otherwise the lower tree is outside the bounds
  and doesn't need to be traversed.  If the upper limit is greater than or
  equal to the current key then the left subtree evaluation doesn't have to
  check the upper bound as it is completely below it. */

  if (ComparisonLower < 0) /* If lower bound is less than current key. */
  {
    if (!AVLDupRecursiveRangeIterate (ArgsPntr, CurrentNode->smallerChildPntr,
    TestLowerBound, (ComparisonUpper >= 0) ? false : TestUpperBound))
      return false; /* The user requested an early abort of the iteration. */
  }

  /* Output the current node, if it is between lower and upper bounds.  Well,
  maybe not - if the user doesn't want to output nodes equal to the start and
  end bounds. */

  if ((ComparisonLower < 0 ||
  (ComparisonLower == 0 && ArgsPntr->includeThingEqualToStart)) &&
  (ComparisonUpper > 0 ||
  (ComparisonUpper == 0 && ArgsPntr->includeThingEqualToEnd)))
  {
    if (!ArgsPntr->iterationCallback (&CurrentNode->key, &CurrentNode->value,
    ArgsPntr->extraUserData))
      return false; /* The user requested an early abort of the iteration. */
  }

  /* Examine the right subtree.  The upper limit of the range has to be larger
  than the current key otherwise nothing needs to be done.  If the lower limit
  is less than or equal to the current key then no lower limit checks need to
  be done for the subtree. */

  if (ComparisonUpper > 0)
  {
    if (!AVLDupRecursiveRangeIterate (ArgsPntr, CurrentNode->largerChildPntr,
    (ComparisonLower <= 0) ? false : TestLowerBound, TestUpperBound))
      return false; /* The user requested an early abort of the iteration. */
  }

  return true; /* Got successfully to the end of this subtree iteration. */
}



/* Internal recursive function for iterating backwards over a subtree without
doing any tests.  Even assumes that the entry node is non-NULL.  Returns TRUE
if successful, FALSE if the user callback aborted the iteration. */

static bool AVLDupRecursiveSimpleBackwardsIterate (
  NonRecursiveArgumentsPointer ArgsPntr,
  AVLDupNodePointer CurrentNode)
{
  /* Output right subtree first, since we are doing it in descending order. */

  if (CurrentNode->largerChildPntr != NULL)
  {
    if (!AVLDupRecursiveSimpleBackwardsIterate (ArgsPntr,
    CurrentNode->largerChildPntr))
      return false; /* User aborted. */
  }

  /* Output the middle node. */

  if (!ArgsPntr->iterationCallback (&CurrentNode->key, &CurrentNode->value,
  ArgsPntr->extraUserData))
    return false; /* The user requested an early abort of the iteration. */

  /* Output left subtree last. */

  if (CurrentNode->smallerChildPntr != NULL)
  {
    if (!AVLDupRecursiveSimpleBackwardsIterate (ArgsPntr,
    CurrentNode->smallerChildPntr))
      return false; /* User aborted. */
  }

  return true;
}



/* Recursively iterate over the tree in descending order, cutting off
traversals which don't fit in the given range of keys.  Returns TRUE if it got
all the way back to the beginning of the range.  If TestLowerBound is TRUE then
tests will be done against userKey1 (the lower bound key).  If it is FALSE then
we will assume that the node and all its children are greater than or equal to
userKey1 and avoid the test.  Similarly if TestUpperBound is FALSE then we
assume everything is less than or equal to the upper bound.  If both are false,
we revert to a simple tree traversal and do no tests. */

static bool AVLDupRecursiveRangeBackwardsIterate (
  NonRecursiveArgumentsPointer ArgsPntr,
  AVLDupNodePointer CurrentNode,
  bool TestLowerBound,
  bool TestUpperBound)
{
  int ComparisonLower;
  int ComparisonUpper;

  if (CurrentNode == NULL)
    return true; /* Successfully finished iterating the NULL tree. */

  if (!(TestLowerBound || TestUpperBound))
    return AVLDupRecursiveSimpleBackwardsIterate (ArgsPntr, CurrentNode);

  /* For convenience in understanding the code, both high and low comparisons
  are done as (BoundsLimit - CurrentNodeKey).  Meaning ComparisonResult is
  less than zero for key bigger than the bounds limit, and so on. */

  if (TestLowerBound)
  {
    ComparisonLower = ArgsPntr->keyComparisonFunctionPntr (
    &ArgsPntr->userKey1, &CurrentNode->key);

    if (ComparisonLower == 0) /* Equal keys, use the value to decide. */
    {
      if (ArgsPntr->howToCompareUserValue1 == 0)
      {
        if (ArgsPntr->userValue1 < CurrentNode->value)
          ComparisonLower = -1;
        else if (ArgsPntr->userValue1 == CurrentNode->value)
          ComparisonLower = 0;
        else
          ComparisonLower = 1;
      }
      else /* Lower bound is plus or minus infinity as requested. */
        ComparisonLower = ArgsPntr->howToCompareUserValue1;
    }
  }
  else /* No comparison, current is always larger. */
    ComparisonLower = -1;

  if (TestUpperBound)
  {
    ComparisonUpper = ArgsPntr->keyComparisonFunctionPntr (
    &ArgsPntr->userKey2, &CurrentNode->key);

    if (ComparisonUpper == 0) /* Equal keys, use the value to decide. */
    {
      if (ArgsPntr->howToCompareUserValue2 == 0)
      {
        if (ArgsPntr->userValue2 < CurrentNode->value)
          ComparisonUpper = -1;
        else if (ArgsPntr->userValue2 == CurrentNode->value)
          ComparisonUpper = 0;
        else
          ComparisonUpper = 1;
      }
      else /* Upper bound is plus or minus infinity as requested. */
        ComparisonUpper = ArgsPntr->howToCompareUserValue2;
    }
  }
  else /* No comparison, current is always smaller. */
    ComparisonUpper = 1;

  /* Examine the right subtree.  The upper limit of the range has to be larger
  than the current key otherwise nothing needs to be done.  If the lower limit
  is less than or equal to the current key then no lower limit checks need to
  be done for the subtree. */

  if (ComparisonUpper > 0)
  {
    if (!AVLDupRecursiveRangeBackwardsIterate (ArgsPntr,
    CurrentNode->largerChildPntr,
    (ComparisonLower <= 0) ? false : TestLowerBound, TestUpperBound))
      return false; /* The user requested an early abort of the iteration. */
  }

  /* Output the current node, if it is between lower and upper bounds.  Well,
  maybe not - if the user doesn't want to output nodes equal to the start and
  end bounds. */

  if ((ComparisonLower < 0 ||
  (ComparisonLower == 0 && ArgsPntr->includeThingEqualToStart)) &&
  (ComparisonUpper > 0 ||
  (ComparisonUpper == 0 && ArgsPntr->includeThingEqualToEnd)))
  {
    if (!ArgsPntr->iterationCallback (&CurrentNode->key, &CurrentNode->value,
    ArgsPntr->extraUserData))
      return false; /* The user requested an early abort of the iteration. */
  }

  /* Examine the left subtree.  The lower limit of the range has to be less
  than the current node's key otherwise the lower tree is outside the bounds
  and doesn't need to be traversed.  If the upper limit is greater than or
  equal to the current key then the left subtree evaluation doesn't have to
  check the upper bound as it is completely below it. */

  if (ComparisonLower < 0) /* If lower bound is less than current key. */
  {
    if (!AVLDupRecursiveRangeBackwardsIterate (ArgsPntr,
    CurrentNode->smallerChildPntr,
    TestLowerBound, (ComparisonUpper >= 0) ? false : TestUpperBound))
      return false; /* The user requested an early abort of the iteration. */
  }

  return true; /* Got successfully to the end of this subtree iteration. */
}



/******************************************************************************
 * This function will call the user provided callback function for every
 * key/value pair in the given range in ascending or descending order,
 * optionally including ones which equal the start and end keys.  AVLDupIterate
 * will return TRUE if it reached the end of the iteration, or FALSE if you
 * stopped it early.
 *
 * If you want to start at the beginning, pass in NULL for StartKeyPntr
 * (StartValue will be ignored in that case).  Otherwise, when specifying a
 * non-NULL key, you can also use HowToCompareStartValue to specify how you
 * want the value of the starting key to be determined, to distinguish betwee
 * several keys with the same starting key and different values.  Of course, if
 * the keys it is comparing are different from the starting key then it won't
 * test the values.  If you want to start with keys with the key equal to the
 * starting key and that have a corresponding value greater than negative
 * infinity, use -1 (basically includes all keys equal to the initial key, no
 * matter what their value is, useful for iterating over all values for a given
 * key).  For keys with values greater than plus infinity, use +1 for
 * HowToCompareStartValue (basically no keys, so the iteration will start at
 * the next higher key).  Finally, with HowToCompareStartValue set to zero, it
 * will test the values against StartValue and use the first key/value which is
 * greater than StartKeyPntr/StartValue.  If IncludeThingEqualToStart is TRUE
 * then that comparison is greater than or equal, not just greater than
 * (doesn't matter when comparing against infinities).
 *
 * Similarly you can use EndKeyPntr, HowToCompareEndValue, EndValue,
 * IncludeThingEqualToEnd parameters to define the ending key.
 *
 * One example: a full tree traversal can be done by passing in NULL for both
 * StartKeyPntr and EndKeyPntr (other settings don't matter).  Another example:
 * for all values for a given key, pass in the same key as StartKeyPntr and
 * EndKeyPntr, set HowToCompareStartValue to -1 (minus infinity for starting
 * value) and HowToCompareEndValue to +1 (plus infinity for ending value).  Yet
 * another example: to resume an iteration, set StartKeyPntr and StartValue to
 * the key and value of the last item returned previously, set
 * HowToCompareStartValue to zero and IncludeThingEqualToStart to FALSE.
 * Usually EndKeyPntr will be NULL so that it continues iterating from the next
 * key/value all the way to the end (or until your callback function decides to
 * stop).
 *
 * While the iteration is proceeding, you shouldn't modify the tree, so other
 * threads only should do read operations.  For that matter, your callback
 * function can't call routines which modify the tree either, or change the
 * key/value items it is passed (that's why you get a const pointer for to the
 * current the key).  If your callback function returns FALSE, then the
 * iteration will be stopped early.  ExtraUserData is also passed to your
 * function, it can be whatever you want it to be (usually used as a pointer to
 * some global variables).
 */

bool AVLDupIterate (
  AVLDupTreePointer TreePntr,
  bool AscendingOrder,
  AVLDupThingConstPointer StartKeyPntr,
  int HowToCompareStartValue,
  unsigned int StartValue,
  bool IncludeThingEqualToStart,
  AVLDupThingConstPointer EndKeyPntr,
  int HowToCompareEndValue,
  unsigned int EndValue,
  bool IncludeThingEqualToEnd,
  AVLDupIterationCallbackFunctionPointer CallbackFunctionPntr,
  void *ExtraUserData)
{
  NonRecursiveArgumentsRecord Arguments;
  bool                        Successful;

  if (TreePntr == NULL || CallbackFunctionPntr == NULL)
    return false;

  Arguments.keyType = TreePntr->keyType;
  Arguments.keyComparisonFunctionPntr = TreePntr->keyComparisonFunctionPntr;
  Arguments.includeThingEqualToStart = IncludeThingEqualToStart;
  Arguments.includeThingEqualToEnd = IncludeThingEqualToEnd;
  Arguments.iterationCallback = CallbackFunctionPntr;
  Arguments.extraUserData = ExtraUserData;
  Arguments.mallocPntr = TreePntr->mallocPntr;
  Arguments.freePntr = TreePntr->freePntr;
  Arguments.poolID = TreePntr->poolID;

  /* Copy the starting key and value, if present, to our semi-global data. */

  if (StartKeyPntr != NULL)
  {
    Arguments.userKey1 = *StartKeyPntr;
    Arguments.userValue1 = StartValue;
    Arguments.howToCompareUserValue1 = HowToCompareStartValue;
  }

  /* Copy the ending key and value, if present. */

  if (EndKeyPntr != NULL)
  {
    Arguments.userKey2 = *EndKeyPntr;
    Arguments.userValue2 = EndValue;
    Arguments.howToCompareUserValue2 = HowToCompareEndValue;
  }

  /* Start off the big recursive iteration.  Forwards or backwards depending on
  what the caller wanted. */

  if (AscendingOrder)
    Successful = AVLDupRecursiveRangeIterate (&Arguments,
      TreePntr->rootPntr, StartKeyPntr != NULL, EndKeyPntr != NULL);
  else
    Successful = AVLDupRecursiveRangeBackwardsIterate (&Arguments,
      TreePntr->rootPntr, StartKeyPntr != NULL, EndKeyPntr != NULL);

  return Successful;
}



/* An internal function for use by AVLDupFind to find a particular key.  Once
it's found, it sets the return value (pointed to by ExtraData) to the matching
value, and aborts the iteration. */

static bool CallbackForAVLDupFind (
  AVLDupThingConstPointer KeyPntr,
  const unsigned int *ValuePntr,
  void *ExtraData)
{
  unsigned int *ResultPntr;

  ResultPntr = ExtraData;
  *ResultPntr = *ValuePntr;
  return false; /* Stop iterating. */
}



/******************************************************************************
 * A convenience function for looking up the first value which has the given
 * key.  Returns zero if it can't find the value.  So, it's best to not use
 * NULL/zero as a value when you add key/value pairs!
 */

unsigned int AVLDupFind (
  AVLDupTreePointer TreePntr,
  AVLDupThingConstPointer KeyPntr)
{
  unsigned int Result;

  if (TreePntr == NULL || KeyPntr == NULL)
    return 0;

  Result = 0;

  AVLDupIterate (TreePntr,
    true /* AscendingOrder */,
    KeyPntr /* StartKeyPntr */,
    -1 /* HowToCompareStartValue */,
    0 /* StartValue */,
    true /* IncludeThingEqualToStart */,
    KeyPntr /* EndKeyPntr */,
    +1 /* HowToCompareEndValue */,
    0 /* EndValue */,
    true /* IncludeThingEqualToEnd */,
    CallbackForAVLDupFind,
    &Result /* ExtraUserData */);

  return Result;
}



/******************************************************************************
 * A convenience function for seeing if a particular key and value pair is in
 * the tree.  Returns TRUE if present, FALSE in all other situations.
 */

bool AVLDupKeyAndValueExists (
  AVLDupTreePointer TreePntr,
  AVLDupThingConstPointer KeyPntr,
  unsigned int Value)
{
  unsigned int UnusedResult;

  if (TreePntr == NULL || KeyPntr == NULL)
    return false;

  UnusedResult = 0;

  return !AVLDupIterate (TreePntr,
    true /* AscendingOrder */,
    KeyPntr /* StartKeyPntr */,
    0 /* HowToCompareStartValue */,
    Value /* StartValue */,
    true /* IncludeThingEqualToStart */,
    KeyPntr /* EndKeyPntr */,
    0 /* HowToCompareEndValue */,
    Value /* EndValue */,
    true /* IncludeThingEqualToEnd */,
    CallbackForAVLDupFind,
    &UnusedResult /* ExtraUserData */);
}



/* An internal function for use by AVLDupFindNext or AVLDupFindPrevious to find
the next larger/smaller key (ignoring values).  Once it's found, it sets the
return parameters (pointed to by ExtraData) to the matching key/value, and
aborts the iteration. */

typedef struct AVLDupFindNextStruct
{
  AVLDupThingConstPointer *nextKeyPntrPntr;
  unsigned int             nextValue;
} AVLDupFindNextRecord, *AVLDupFindNextPointer;

static bool CallbackForAVLDupFindNextOrPrevious (
  AVLDupThingConstPointer KeyPntr,
  const unsigned int *ValuePntr,
  void *ExtraData)
{
  AVLDupFindNextPointer ResultPntr = ExtraData;

  *(ResultPntr->nextKeyPntrPntr) = KeyPntr;
  ResultPntr->nextValue = *ValuePntr;

  return false; /* Stop iterating. */
}



/******************************************************************************
 * A convenience function for finding the next larger key (and associated
 * value) when given an existing key to start looking from (ignores duplicates
 * of the starting key when trying to find the next key).  Returns the value if
 * it found it, and updates the pointer provided for the key (doesn't
 * deallocate or overwrite your passed in starting key).  Returns zero and does
 * nothing to your pointer if it can't find a larger key.
 */

unsigned int AVLDupFindNext (
  AVLDupTreePointer TreePntr,
  AVLDupThingConstPointer *KeyPntrPntr)
{
  AVLDupFindNextRecord  ResultStorage;

  if (TreePntr == NULL || KeyPntrPntr == NULL)
    return 0;

  ResultStorage.nextKeyPntrPntr = KeyPntrPntr;
  ResultStorage.nextValue = 0;

  AVLDupIterate (TreePntr,
    true /* AscendingOrder */,
    (AVLDupThingPointer) *KeyPntrPntr /* StartKeyPntr */,
    +1 /* HowToCompareStartValue */,
    0 /* StartValue */,
    false /* IncludeThingEqualToStart */,
    NULL /* EndKeyPntr */,
    +1 /* HowToCompareEndValue */,
    0 /* EndValue */,
    false /* IncludeThingEqualToEnd */,
    CallbackForAVLDupFindNextOrPrevious,
    &ResultStorage /* ExtraUserData */);

  return ResultStorage.nextValue;
}



/******************************************************************************
 * A convenience function for finding the previous key (and associated value)
 * when given an existing key to start looking from (ignores duplicates of the
 * starting key when trying to find the previous key).  Returns the value if it
 * found it, and updates the pointer provided for the key (doesn't deallocate
 * or overwrite your passed in starting key).  Returns zero and does nothing to
 * your pointer if it doesn't have any prior key/values.
 */

unsigned int AVLDupFindPrevious (
  AVLDupTreePointer TreePntr,
  AVLDupThingConstPointer *KeyPntrPntr)
{
  AVLDupFindNextRecord  ResultStorage;

  if (TreePntr == NULL || KeyPntrPntr == NULL)
    return 0;

  ResultStorage.nextKeyPntrPntr = KeyPntrPntr;
  ResultStorage.nextValue = 0;

  AVLDupIterate (TreePntr,
    false /* AscendingOrder */,
    NULL /* StartKeyPntr */,
    -1 /* HowToCompareStartValue */,
    0 /* StartValue */,
    false /* IncludeThingEqualToStart */,
    (AVLDupThingPointer) *KeyPntrPntr /* EndKeyPntr */,
    -1 /* HowToCompareEndValue */,
    0 /* EndValue */,
    false /* IncludeThingEqualToEnd */,
    CallbackForAVLDupFindNextOrPrevious,
    &ResultStorage /* ExtraUserData */);

  return ResultStorage.nextValue;
}



/* An internal function for use by AVLDupChangeValueInPlace to change the value
for a particular key. */

typedef struct AVLDupChangeValueInPlaceStruct
{
  unsigned int      newValue;
  unsigned int      oldValue;
} AVLDupChangeValueInPlaceRecord, *AVLDupChangeValueInPlacePointer;

static bool CallbackForAVLDupChangeValueInPlace (
  AVLDupThingConstPointer KeyPntr,
  const unsigned int *ValuePntr,
  void *ExtraData)
{
  AVLDupChangeValueInPlacePointer ArgsPntr = ExtraData;

  if (*ValuePntr != ArgsPntr->oldValue)
    return false;

  /* Override the const value pointer since we want to replace the value,
  unsafe though that may be. */

  * (unsigned int *) ValuePntr = ArgsPntr->newValue;

  return true; /* Successful, continue iterating (will stop right away). */
}



/******************************************************************************
 * Change the value for a given key.  It is done in place (just changing the
 * node's value field, not rebalancing the tree) for speed and to guarantee it
 * will always work (no allocations or deallocations are done).  That means it
 * won't work properly if there is another node with the same key, because with
 * equal keys the values are used to order the nodes in the tree.  In other
 * words, it is safe to use this function only in trees with unique keys.  You
 * need to specify both the key and old value to identify the existing node
 * (mostly as a bug check).  Returns TRUE if successful, FALSE if it doesn't
 * find the key or the old value doesn't match the current value.
 */

bool AVLDupChangeValueInPlace (
  AVLDupTreePointer TreePntr,
  AVLDupThingConstPointer KeyPntr,
  unsigned int OldValue,
  unsigned int NewValue)
{
  AVLDupChangeValueInPlaceRecord Args;

  if (TreePntr == NULL || KeyPntr == NULL)
    return false;

  Args.newValue = NewValue;
  Args.oldValue = OldValue;

  return AVLDupIterate (TreePntr,
    true /* AscendingOrder */,
    KeyPntr /* StartKeyPntr */,
    0 /* HowToCompareStartValue */,
    OldValue /* StartValue */,
    true /* IncludeThingEqualToStart */,
    KeyPntr /* EndKeyPntr */,
    0 /* HowToCompareEndValue */,
    OldValue /* EndValue */,
    true /* IncludeThingEqualToEnd */,
    CallbackForAVLDupChangeValueInPlace,
    &Args /* ExtraUserData */);
}
